#!/usr/bin/env python3

from math import floor
import numpy as np
from .parser_gau import flat_list
import .seminario_module as sem_mod

def solve_2Dsys(n1, n2, x, grad, k_tors):
    AM = np.array([[n1 * 0.5* np.sin(n1*x), -n2 *0.5* np.sin(n2*x) ],
                   [n1*n1 * 0.5* np.cos(n1*x), -n2*n2 * 0.5* np.cos(n2*x) ]])
    b = np.array([-grad, k_tors])
    coeffs = np.linalg.solve(AM, b)

    return abs(coeffs)

def flatList_to_2Darray(my_list):
    tmp_arr = np.array(my_list)
    array_2d = np.reshape(tmp_arr, ((tmp_arr.shape[0], 1)) )
    return array_2d
    

def avg_dups(genes, values):
    """ 
    Average values on the basis of 
    their genes
    """
    folded, indices, counts = np.unique(genes, return_inverse=True, return_counts=True)

    output = np.zeros((folded.shape[0], values.shape[1]))
    np.add.at(output, indices, values)
    output /= counts[:, np.newaxis]

    return folded, output


def set_bonds(coords, hess, type_list,
              bond_list, k_bonds, mdin, mdout):
    """
    Order Bond Force Constants
    """
    bond_length_list = []
    bond_type_list = []
    for m, k in enumerate(bond_list):
        i = k[0] -1 
        j = k[1] -1
        diff_AB = coords[i,:] - coords[j,:]
        r_AB = np.linalg.norm(diff_AB)
        bond_length_list.append(r_AB)
        bond_type_list.append(type_list[i] + ' ' + type_list[j]) 
        if mdin == 'sem':            # Triggers on Seminario 
            k_bonds[m] = sem_mod.get_ModSem_FcBonds(i, j, diff_AB, r_AB, hess)

    bond_type_list = flat_list(bond_type_list)

    # Average over values if duplicates found,
    # return all of'em
    if mdout == 'mean':
        tmp_arr = np.array(bond_length_list)
        bond_length_2d = np.reshape(tmp_arr, ((tmp_arr.shape[0], 1)) )
        folded, out_1 = avg_dups(bond_type_list, bond_length_2d)
        bond_type_unique = list(folded)
        bond_length_mean = np.reshape(out_1, (out_1.shape[0]))

        k_bonds_2d = np.reshape(k_bonds, ((k_bonds.shape[0], 1)) )
        _, out_2 = avg_dups(bond_type_list, k_bonds_2d)
        k_bonds_mean = np.reshape(out_2, (out_2.shape[0]))
        return bond_type_unique, bond_length_mean, k_bonds_mean
    elif mdout == 'all':
        return bond_type_list, bond_length_list, k_bonds
        
def set_angles(coords, hess, type_list,
              angle_list, k_angles, mdin, mdout):
    """
    Order Angles Force Constants
    """
    angle_length_list = []
    angle_type_list = []
    for m, p in enumerate(angle_list):
        i = p[0] - 1 
        j = p[1] - 1
        k = p[2] - 1
        diff_AB = coords[i,:] - coords[j,:]
        diff_BC = coords[j,:] - coords[k,:]
        r_AB = np.linalg.norm(diff_AB)
        r_BC = np.linalg.norm(diff_BC)
        u_AB = diff_AB / r_AB
        u_BC = diff_BC / r_BC
        cos_theta = np.dot(u_AB, u_BC)
        theta = np.arccos(cos_theta)
        theta = 180 - theta * 180 / np.pi
        angle_length_list.append(theta)
        angle_type_list.append(type_list[i] + ' ' + type_list[j] +  \
                               ' ' + type_list[k]) 
        if mdin == 'sem':            # Triggers on ModSeminario 
           k_angles[m] = sem_mod.get_ModSem_FcAngles(i, j, k, r_AB, r_BC, \
                                                     u_AB, u_BC, hess)

    angle_type_list = flat_list(angle_type_list)

    # # Average over values if duplicates found,
    # # return all of'em
    if mdout == 'mean':
        tmp_arr = np.array(angle_length_list)
        angle_length_2d = np.reshape(tmp_arr, ((tmp_arr.shape[0], 1)) )
        folded, out_1 = avg_dups(angle_type_list, angle_length_2d)
        angle_type_unique = list(folded)
        angle_length_mean = np.reshape(out_1, (out_1.shape[0]))

        k_angles_2d = np.reshape(k_angles, ((k_angles.shape[0], 1)) )
        _, out_2 = avg_dups(angle_type_list, k_angles_2d)
        k_bonds_mean = np.reshape(out_2, (out_2.shape[0]))

        return angle_type_unique, angle_length_mean, k_bonds_mean
    elif mdout == 'all':
        return angle_type_list, angle_length_list, k_angles
        
        

def set_torsion(coords, type_list, tors_list, \
                k_tors, grad, mdout):
    """
    Order Dihedral Force Constants
    """
    center_list = [ x[1:3] for x in tors_list]
    hybrid_list = [ center_list.count(x) for x in center_list]
    v1_eq = [0] * len(tors_list)
    v2_eq = [0] * len(tors_list)
    v3_eq = [0] * len(tors_list)
    tors_length_list = []
    tors_type_list = []
    phase = np.zeros((len(tors_list), 4))
    for m, p in enumerate(tors_list):
            i = p[0] - 1 
            j = p[1] - 1
            k = p[2] - 1
            l = p[3] - 1
            diff_AB = coords[i,:] - coords[j,:]
            diff_BC = coords[j,:] - coords[k,:]
            diff_CD = coords[k,:] - coords[l,:]
            u_ABC = np.cross(diff_AB, diff_BC)
            u_BCD = np.cross(diff_BC, diff_CD)
            u_ABC_n = np.linalg.norm(u_ABC)
            u_BCD_n = np.linalg.norm(u_BCD)
            cos_phi = np.dot(u_ABC, u_BCD) / ( u_ABC_n * u_BCD_n)
            # Floor cos_phi if out of [-1,1] range
            if cos_phi > 1.0 or cos_phi < -1.0:
                cos_phi = float(floor(cos_phi))
            phi = np.arccos(cos_phi)
            phi_deg = phi * 180 / np.pi
            tors_length_list.append(phi_deg)
            my_str = type_list[i] + ' ' + type_list[j] + \
                     ' ' + type_list[k] + ' ' + type_list[l]
            tors_type_list.append(my_str) 
            
            # Compute Torsional Force Constant 
            # on the basis  of torsion periodicity: sp3,sp2,sp
            if hybrid_list[m] == 9:           
                del_1 = abs(phi_deg - 60.)
                del_2 = abs(phi_deg - 180.)
                eps = 5.0
                if del_1 < eps or del_2 < eps:
                     n, d = 3.0, 1.0
                     v3 = abs( - (d * np.abs(k_tors[m]))/(n*n* np.cos(n*phi_deg)) )
                    #  print(k_tors[m], v3, n)
                     if v3 > 5.0:                         # If force constants too stiff
                        v3_eq[m] = np.exp(-1.4/v3)*1.4     # use AMBER X-C-C-X values
                     else:
                        v3_eq[m] = v3
                    #  print(f' {phi_deg:.2f}  {v3_eq[m]:.2f} {hybrid_list[m]}')
                else:
                    v2, v3 = solve_2Dsys(2, 3, phi, grad[m], k_tors[m])
                    if v2 > 5.0 or v3 > 5.0:
                        v2_eq[m] = 0.0
                        v3_eq[m] = 1.4
                    else:
                        v2_eq[m] = v2
                        v3_eq[m] = v3
                    # print(f' {phi_deg:.2f} {v2_eq[m]:.2f} {v3_eq[m]:.2f} {hybrid_list[m]}')
            elif hybrid_list[m] == 4 or hybrid_list[m] == 2:          # sp2-sp2 bonds
                del_1 = abs(phi_deg - 0.)
                del_2 = abs(phi_deg - 180.)
                eps = 5.0
                phase[m, 1] = 180
                if del_1 < eps or del_2 < eps:
                     n, d = 2.0, 1.0
                     v2 = abs( (d * np.abs(k_tors[m]))/(n*n* np.cos(n*phi_deg)) )
                    #  print(k_tors[m], v2, n)
                    #  if v2 > 30 or 0 < v2 < 9.0:
                     if v2 > 30:
                        v2_eq[m] = np.exp(-30./v2)*v2       # Reduce to a full C=C
                        # print(v2_eq[m])
                        # print("")
                     elif 0 < v2 < 14.5:
                          v2_eq[m] = np.exp(-v2/30.)*30. - np.exp(-v2/14.5)*14.5   # Scale for full C=C & 
                     else:                                 # Increase to a partial C=C
                        v2_eq[m] = v2
                    #  print(f'{phase[m, 1]}  {v2:.2f} {hybrid_list[m]}')
                else:
                    v1, v2 = solve_2Dsys(1, 2, phi, grad[m], k_tors[m])
                    if v1 > 30 or v2 > 30:
                        v1_eq[m] = 0.0
                        v2_eq[m] = np.exp(-v2/30.)*30.
                    else:
                        v1_eq[m] = v1
                        v2_eq[m] = v2
                    # print(f' {phi_deg:.2f} {v1:.2f} {v2:.2f} {hybrid_list[m]}')
            else:
                n, d = 2.0, 1.0
                v2 = abs( -2* (d * k_tors[m])/(n*n* np.cos(n*phi_deg)) )
                if v2 > 30:
                   v2_eq[m] = 14.5
                else:
                   v2_eq[m] = v2
                # print(f' {phi_deg:.2f} {v2:.2f} {hybrid_list[m]}')
    
    tors_type_list = flat_list(tors_type_list)
    phase = phase.astype(int)
    
    # for i in range(len(tors_list)):
    #     print(f'{hybrid_list[i]} {v1_eq[i]:.1f}  {v2_eq[i]:.1f}  {v3_eq[i]:.1f}')
    # # Average over values if duplicates found,
    # # return all of'em
    if mdout == 'mean':
        tors_length_2d = flatList_to_2Darray(tors_length_list)
        folded, out_0 = avg_dups(tors_type_list, tors_length_2d)
        tors_type_unique = list(folded)
        tors_length_mean = np.reshape(out_0, (out_0.shape[0]))

        hybrid_2d = flatList_to_2Darray(hybrid_list)
        _, out_1 = avg_dups(tors_type_list, hybrid_2d)
        tors_type_unique = list(folded)
        hybrid_mean = np.reshape(out_1, (out_1.shape[0]))

        v1_eq_2d = flatList_to_2Darray(v1_eq)
        _, out_1 = avg_dups(tors_type_list, v1_eq_2d)
        v1_eq_mean = np.reshape(out_1, (out_1.shape[0]))

        v2_eq_2d = flatList_to_2Darray(v2_eq)
        _, out_2 = avg_dups(tors_type_list, v2_eq_2d)
        v2_eq_mean = np.reshape(out_2, (out_2.shape[0]))

        v3_eq_2d = flatList_to_2Darray(v3_eq)
        _, out_3 = avg_dups(tors_type_list, v3_eq_2d)
        v3_eq_mean = np.reshape(out_3, (out_3.shape[0]))

        return tors_type_unique, v1_eq_mean, \
               v2_eq_mean, v3_eq_mean,  tors_length_mean, phase, hybrid_mean
    elif mdout == 'all':
        return tors_type_list, v1_eq, \
               v2_eq, v3_eq, tors_length_list, phase, hybrid_list
        





#!/usr/bin/env python3 

import sys
import json
import os

def print_mm(fname, elements, xyz, Natm, atom_types, force):
# opt=(maxcycle=300)
    header = """%mem=1GB
%nprocshared=1
%chk={CHK}
#p Amber=(SoftFirst,Print) geom=nocrowd nosymm
   
Title
  
0 1
"""
    file_mm_list =[]
    for i in range(0, len(xyz), Natm):
            j = i + Natm
            f = os.path.splitext(fname)[0][:-2]
            fout = f + 'mm_' + str(i) + '.gjf'
            file_mm_list.append(str(fout))
            with open(fout, 'w') as fopen:
                 fopen.write(header.format(CHK=fout[:-3]+'chk'))
                #  fopen.write(f' {Natm}\n')
                #  fopen.write('\n')
                 for m, p, l in zip(elements[i:j], atom_types, xyz[i:j]):
                     s1 = '  '.join(str(x) for x in p)
                     s2 = '  '.join(str(x) for x in l)
                     fopen.write(f'{m}-{s1}      {s2} \n')
                 fopen.write(f'\n')
                 for x in force:
                    l = ' '.join(x)
                    fopen.write(f'{l}\n')
                 fopen.write(f'\n')
    return file_mm_list



def read_optfile(fname):
    """
    Read the option file. expected in json style
    """
    if not os.path.exists(fname):
        raise FileNotFoundError('Missing JSON file')
    with open(fname, 'r')  as fopen:
        data = json.load(fopen)
    for i in ['atom2type', 'force_file', 'log_file']:
        if not os.path.exists(data['files'][i]):
            raise FileNotFoundError(f'Missing {i} file')
    return data

def read_txt_info(fname):
    """
    Parse txt files
    """
    # data = ""
    data = []
    with open(fname, 'r', encoding='UTF-8') as fopen:
        for line in fopen:
            data.append(line.split())
            # data += str(line)
    return data

def read_log(logfile):
    all_lines = []  
    with open (logfile, 'r') as file:                          
        for each_line in file:
            all_lines.append(each_line.strip())
    
    i = 0
    k = 0
    data = []
    for line in range(len(all_lines)):
        if "Optimized Parameters" in all_lines[line]:
            k = k + 1
            start = line
            # print(k, start, all_lines[line])
            for s in range(start, len(all_lines)): 
                j = 0                              
                if "Input orientation:" in all_lines[s]:
                    start2 = s
                    i = i + 1
                    # print(k, i, start2, all_lines[start2])
                    for e in range(start2 + 5, len(all_lines)):
                                  if "----" in all_lines[e]:
                                     end = e
                                     break
                                  j = j + 1
                                  data.append(all_lines[e])
                    break
    
    
    atomic_number = {
        '1':'H', '2':'He', '3':'Li', '4':'Be', '5':'B', '6':'C', '7':'N', '8':'O', '9':'F', '10':'Ne',
        '11':'Na', '12':'Mg', '13':'Al', '14':'Si','15':'P','16':'S','17':'Cl','18':'Ar','19':'K',
        '20':'Ca', "21" : "Sc" , "22" : "Ti" , "23" : "V"  , "24" : "Cr" , "25" : "Mn", "26" : "Fe" ,
        "27": "Co" , "28" : "Ni" , "29" : "Cu" , "30" : "Zn", "31" : "Ga" , "32" : "Ge" , "33" : "As" ,
        "34": "Se" , "35" : "Br", "36" : "Kr" , "37" : "Rb" , "38" : "Sr" , "39" : "Y"  , "40" : "Zr",
        "41": "Nb" , "42" : "Mo" , "43" : "Tc" , "44" : "Ru" , "45" : "Rh", "46" : "Pd" , "47" : "Ag" ,
        "48": "Cd" , "49" : "In" , "50" : "Sn", "51" : "Sb" , "52" : "Te" , "53" : "I"  , "54" : "Xe" ,
        "55": "Cs", "56" : "Ba" , "57" : "La" , "58" : "Ce" , "59" : "Pr" , "60" : "Nd", "61" : "Pm" ,
        "62": "Sm", "63" : "Eu" , "64" : "Gd" , "65" : "Tb", "66" : "Dy" , "67" : "Ho" , "68" : "Er" ,
        "69": "Tm", "70" : "Yb", "71" : "Lu" , "72" : "Hf" , "73" : "Ta" , "74" : "W"  , "75" : "Re",
        "76": "Os", "77" : "Ir" , "78" : "Pt" , "79" : "Au" , "80" : "Hg", "81" : "Tl" , "82" : "Pb" ,
        "83": "Bi", "84" : "Po" , "85" : "At", "86" : "Rn" , "87" : "Fr" , "88" : "Ra" , "89" : "Ac" ,
        "90": "Th", "91" : "Pa" , "92" : "U"  , "93" : "Np" , "94" : "Pu" , "95" : "Am", "96" : "Cm" ,
        "97": "Bk", "98" : "Cf" , "99" : "Es" ,"100" : "Fm", "101": "Md" ,"102" : "No" ,"103" : "Lr" ,
        "104": "Rf","105" : "Db", "106": "Sg" ,"107" : "Bh" ,"108" : "Hs" ,"109" : "Mt" ,"110" : "Ds",
        "111": "Rg","112" : "Uub","113" : "Uut","114" : "Uuq","115" : "Uup", "116": "Uuh","117" : "Uus","118" : "Uuo"
        }
    
    # Build Ele, XYZ
    Natm = j                                                       
    elements = []
    xyz_angs =[]
    for line in data:
        words = line.split()
        elements.append(atomic_number.get(words[1]))
        xyz_angs.append(words[3:])

    return elements, xyz_angs, Natm

def main():
    fname = sys.argv[1]
    data = read_optfile(fname)
    logfile = str(data["files"]["log_file"]) 
    ele, xyz, Natm = read_log(logfile)
    ffs = read_txt_info(data["files"]["force_file"])          #  Reading in force field
    atom_types = read_txt_info(data["files"]["atom2type"])         #  Reading in atom types
    print_mm(fname, ele, xyz, Natm, atom_types, ffs)
    

if __name__ == "__main__":
    main()
  






#!/usr/bin/env python3 


import argparse
import numpy as np
import pandas as pd

# def read_gjf(file_gjf):
#     match = "Variables"
#     splitted_line = []
#     with open(file_gjf, 'r' ) as f:
#         for line in f:
#             if line[:9] == match:
#                 break
#         for line in f:
#             splitted_line.append(line.split())
                
#     f.close()
#     for i in splitted_line:
#         for j in i:
#             if j == 'S':
#                 name_ang, start_ang, N_step, size_step = i[0], float(i[1]), int(i[3]), float(i[4])
#     f.close()
#     name_ang_new = name_ang.replace("=", '')
#     return name_ang_new, start_ang, N_step, size_step

def build_parser():
    """
    Build options for parser
    """
    parser = argparse.ArgumentParser()
    txt = 'The type of file to be opened: qm or mm'
    parser.add_argument('-t', type=str, default='qm',
                         help=txt)
    txt = "Scan log file"
    parser.add_argument('-f', type=str, help=txt)
    parser.add_argument('-o', type=str, help='output file', default='output.csv')
    return parser


def read_log(file_log, ftype):
     line_0 = []
     line_1 = []
     splitted_line = []
     match_0 = " The following ModRedundant input section has been read:"
     match_01 = "       Variables:"
     match_1 = " Summary of Optimized Potential Surface Scan"
     match_2 = "Eigenvalues"
     match_3 = " ! Name  Definition              Value          Derivative Info."
     match_4 = " ---"

     # Get Angle step size:
     # If file QM.log search for match_0
     #
     with open(file_log, 'r' ) as f:
        if ftype == 'qm':                            # If log file QM
            for line in f:
                 if line[:56] == match_0:
                     break
            line_0.append(f.readline().split())
            step_ang = float(line_0[0][7])
    
           # Get Starting Angle
            for line in f:
                if line[:64] == match_3:
                    break
            f.readline()
            for line in f:
                if line[:4] == match_4:
                    break
                line_1.append(line.split())
            for i, x in enumerate(line_1):
                if len(x) == 6:
                    angle = float(x[3])
        else:                                   # If log file MM
            for line in f:
                if line[:17] == match_01:
                     break
            for line in f:
                if line == " \n":
                    break
                line_0.append(line.split())
            for i in line_0:
                if len(i) == 5:
                   angle = float(i[1])
                   step_ang = float(i[4])
            # print(angle, step_ang)

        # Get Energies
        for line in f:
            if line[:44] == match_1:
                ipos = line.find('to')
                add = float(line[50:ipos])
                for line in f:
                    ipos = line.find(match_2)
                    if line[ipos:16] == match_2:
                        splitted_line.append(line.split())
     f.close()   

     energy_list = []
     [ energy_list.append(list(map(float,i[2:]))) for i in splitted_line ] 
     energy = [ y for x in energy_list for y in x]
     energy = np.array(energy)
     energy = energy + add
     return step_ang, angle, energy

def main():
    parser = build_parser()
    args = parser.parse_args()

    file_log = args.f
    ftype = args.t
    fout = args.o
    size_step, ang, energy_au = read_log(file_log, ftype)
    file_out = file_log[:-4] + '_scan.csv'
    phi = [ang]
    for i in range(len(energy_au)-1):
        ang = ang + size_step
        phi.append(ang)
    
    energy = list(energy_au)
    df = pd.DataFrame([phi, energy]).T
    # df = pd.DataFrame([phi, energy_au], columns=['Angle', 'Energy'])
    df.to_csv(fout)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import subprocess
import os
import glob
import numpy as np
import .gauScan2com as scan2mm
import .readin_opts as rdin

# import gcutil as gc
# from openbabel import openbabel as ob

def printout_start():
    print("""
    =============================================================
    Author : Emanuele Falbo
    E-Mail : falbo.emanuele@gmail.com
    =============================================================
    SmartField_Torsion : a program to fragment molecules 
                         on the basis of their dihedral angles
                         prepare and compute inputs for subsequent QM and 
                         MM rigid torsional scans
    =============================================================
    """)

#    B3LYP/6-31G* EmpiricalDispersion=GD3  
def print2QM(filename, data, tors_angle, nprocs, qm_method):
    header = """%mem=1GB
%nprocshared={}
%chk={}.chk
#p nosymm geom=nocrowd opt=(modredundant,maxcycle=100) 
  {}

Title

0 1
"""
    s = " ".join(map(str, tors_angle))
    with open(filename,"w") as f:
        f.write(header.format(nprocs, filename[:-4], qm_method))
        for i in data:
            f.write(f"{i[0]}  {' '.join(map(str, i[1:]))}\n")
        f.write('\n')    
        f.write('D ' + str(s) + ' S 10 36.0\n')
        f.write('\n')    
        

def read_XYZ(filename):
    with open(filename, 'r') as f:
        data = []
        N_atoms = int(f.readline().split()[0])      
        f.readline()
        for i in range(0, N_atoms):
               data.append(f.readline().split())
        # ele = [ i[0] for i in data ]
        # XYZ = np.array([ i[1:4] for i in data ], dtype=float)
        # return ele, XYZ
        return data
    

def read_topology(filename):
    with open(filename, 'r') as f:
        bond_list = [] 
        angle_list = []
        tors_list = []
        N_bonds = int(f.readline().split()[0])        
        for i in range(0, N_bonds):
            bond_list.append(f.readline().strip().split())
        
        b_list = [ list(map(int,x)) for x in bond_list ]

        N_angles = int(f.readline().split()[0])       
        for i in range(0, N_angles):
            angle_list.append(f.readline().split())  
        a_list = [ list(map(int,x)) for x in angle_list ] 
        
        N_tors = int(f.readline().split()[0]) 
        for i in range(0, N_tors):
            tors_list.append(f.readline().strip().split())
        
        t_list = [ list(map(int,x)) for x in tors_list ]
        # tors_list = np.array([ i[] for i in tors_list ], dtype=float)
        return t_list   
    

def substitute_strings(indices, strings):
    res = []
    for idx_list in indices:
        substituted = [strings[idx - 1] for idx in idx_list]
        res.append(' '.join(substituted))
    return res

def replace_nth_tors(file_mm, tors_type):
    content =[]
    with open(file_mm, 'r') as f:
        content = [line.split() for line in f]
            
    # Find the index of the '!Torsions' entry
    torsions_index = next((index for index, sublist in enumerate(content) if '!Torsions' in sublist), -1)
    content_tors = content[torsions_index + 1:] if torsions_index != -1 else []
    
    # content_list = content.split('\n')
     # split_content = [item.split() for item in content_tors]
    # Split each item of the list into a sublist based on space delimiter
    for line_idx, line in enumerate(content_tors):
        if line[1:5] == tors_type:
           arr_tmp = np.asarray(content_tors[line_idx][9:13])
           arr_float = arr_tmp.astype(float)
           # Replace elements different from 0.0 with 1.0
           arr_float[arr_float != 0.0] = 0.0
        #    print(line[1:5], tors_type, arr_float, content[line_idx][9:13])
           str_type = arr_float.astype(str)
           content_tors[line_idx][9:13] = list(str_type)
           break
    
    # Write content back to the file
    with open(file_mm, 'w') as f:
        for line in content:
            f.write(' '.join(line) + '\n')
    

     
def main():
    printout_start()
    parser = rdin.commandline_parser3()
    opts = parser.parse_args()
    json_opts = rdin.read_optfile_2(opts.optfile)
    file_xyz = json_opts['files']['file_xyz']
    file_top = json_opts['files']['topol']
    file_atype = json_opts['files']['atom2type']
    file_ff_str = json_opts['files']['force_file']
    nprocs = json_opts['nprocs']
    qm_method = json_opts['method']
    
    data = read_XYZ(file_xyz)
    ele = [ i[0] for i in data ]
    # cords = np.array([ i[1:4] for i in data ], dtype=float)
    tors_list = read_topology(file_top)

    # Building torsional list by element type
    ele_list = []
    for id_1, x in enumerate(tors_list):
        tmp_list = []
        for id_2, y in enumerate(x):
            tmp_list.append(ele[y-1])
        ele_list.append(tmp_list)

    # Explicit tors_unique by element strings
    ele_unique = tuple(set([tuple(i) for i in ele_list]))
    ele_unique_list = [ list(i) for i in ele_unique]
    print(" Torsional Angles by Elements:")
    print(" ----------------------------\n")
    [print(i, ": ", *j) for i, j in enumerate(ele_unique_list)]
    print("")
    
    tmp_center = tuple(set([tuple(i[1:3]) for i in tors_list]))
    print(" Rotation Around Bonds:")
    print(" ----------------------\n") 
    [ print(" {}- {} ".format(i,j)) for i, j in tmp_center ]
    print("")

    # Explicit tors_unique 
    tors_mean_list = []
    for i, x in enumerate(ele_unique_list, start=0):
        for j, y in enumerate(ele_list, start=0):
            if x == y:
               tors_mean_list.append(tors_list[j])
               break

    print(" Atomic indices in Torsional Angles:")
    print(" ---------------------------------------------\n")
    [ print(" {}- {}- {}- {} ".format(i,j, k, l)) for i,j,k,l in tors_mean_list ]
    print("")
    
    # Reading in ff_string and type_charge files
    atype_chg = scan2mm.read_txt_info(file_atype)
    ffs = scan2mm.read_txt_info(file_ff_str)
    atypes = [item[0].split('-')[0] for item in atype_chg]
    
    # Replacing indices with string atype
    tors_tmp = substitute_strings(tors_mean_list, atypes)
    tors_type_list = {idx:x.split(' ') for idx, x in enumerate(tors_tmp)}
    print(" Atomic Type per indices in Torsional Angles:")
    print(" ---------------------------------------------\n")
    [ print(" {}- {}- {}- {} ".format(i,j, k, l)) for i,j,k,l in tors_type_list.values() ]
    print("")
    
    # Writing QM Scan files for each torsional:
    for id, x in enumerate(tors_mean_list):
        fqm = str(id) + '_qm.gjf'
        # fmm = str(id) + '_zmat_mm.gjf'
        print2QM(fqm, data, x, nprocs, qm_method)
    
    GPATH = os.environ.get("g09root") + "/g09"
    file_pattern = '*_qm.gjf*'
    file_qm_list = sorted(glob.glob(file_pattern))      # the order matters
    
    # Calling Gaussian to perfrom Scan on each QM dihedral
    for id, f in enumerate(file_qm_list):
        print(f"Executing Gaussian Scan on {f}")
        file_base = os.path.splitext(f)[0]
        log_file = file_base + '.log'  
        
    # Check if the log files exist 
        if os.path.exists(log_file):
            print(f"Log file {log_file} already exists.\n")
        else:
            process = subprocess.Popen([f"{GPATH}/g09", f], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = process.communicate()  # Wait for the process to finish
            if process.returncode != 0:
                print(f"Error executing Gaussian Scan on {f}. Details: {stderr.decode()}")
            else:
                print(f"Gaussian Scan executed successfully on {f}")
                
        #  Creating MM input geometries for each QM file 
        ele, coords, Natoms = scan2mm.read_log(log_file)
        file_mm_list = scan2mm.print_mm(f, ele, coords, Natoms, atype_chg, ffs)
        
        # Silcening n-th dihedral by replacing with 0.0 barrier terms
        replace_item = tors_type_list[id]
        for idf, file in enumerate(file_mm_list):
            replace_nth_tors(file, replace_item)
        
        # Calling Gaussian to perform Optimization on each QM input for full scan 
        # for jd, fj in enumerate(file_mm_list):
            process = subprocess.Popen([f"{GPATH}/g09", file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = process.communicate()  # Wait for the process to finish
            if process.returncode != 0:
                print(f"Error executing MM Gaussian Optimization on {file}. Details: {stderr.decode()}")
            else:
                print(f"Gaussian Optimization executed successfully on {file}")
    
    
    # Conver to ZMAT by gcutils
    # xyzarr, atomnames = gc.readxyz(xyzfilename)
    # distmat = gc.distance_matrix(xyzarr)
    # gc.write_zmat(xyzarr, distmat, atomnames, rvar=False, avar=False, dvar=True)
    

if __name__ == '__main__':
    main()
    
    



#!/usr/bin/env python3

import subprocess
import os
import .readin_opts as rdin


def print_init():
     print("""
 ======================================================
   Program:      HessFit
   Creator:      Emanuele Falbo, Napoli
   Language:     Python 3.v later
   Description:  The program returns force constants for
                 bonded values and non-bonded parameters
   Mail:         falbo.emanuele@gmail.com 
 =======================================================
     
    """)


def main():
    print_init()
    parser = rdin.commandline_parser3()
    opts = parser.parse_args()
    
    groot = os.environ.get('g09root') 
    print(groot)
    # GPATH = opts.path + "g09"
    if not groot or not os.path.exists(os.path.join(groot, "g09")):
        groot = opts.path
        
    if groot:
        GPATH = os.path.join(groot, "g09")
        
    SM = "HessFit_harmonic.py"
    BS = "build_4Hessfit.py"
    JSON = opts.optfile

    # subprocess.run([BS, JSON, "-path", GPATH], check=True)
    subprocess.run([BS, JSON], check=True)

    for f in ["GauHarm.gjf", "GauNonBon.gjf"]:
        print(f"Executing Gaussian on {f}")
        subprocess.run([f"{GPATH}/g09", f], check=True)

    for f in ["GauHarm.chk", "GauNonBon.chk"]:
        print(f"Formatchecking {f} file")
        subprocess.run([f"{GPATH}/formchk", "-3", f, f"{os.path.splitext(f)[0]}.fchk"], check=True)

    print(f"Executing {SM}")
    subprocess.run(["HessFit_harmonic.py", JSON])
    subprocess.run([f"{GPATH}/g09", "SmartField4gau.gjf"], check=True)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import numpy as np
from numpy.linalg import inv
from scipy import optimize
import matplotlib.pyplot as plt
import pandas as pd


def build_parser():
    """
    Build options for parser
    """
    parser = argparse.ArgumentParser()
    txt = "file containg phi angle, QM, and MM energies in 1st, 2nd, and 3rd columns"
    parser.add_argument('file', type=str, help=txt)
    txt = 'Plot the PES with matplotlib'
    parser.add_argument('--plot', action="store_true", help=txt)
    txt = 'Name of output png file'
    parser.add_argument('--name', default='scan', help=txt)
    return parser


def load_data(fname):
    data = []
    with open(fname,'r') as f:
        for line in f:
            data.append(line.split())
        data = np.array(data, dtype=float)    
    return data

def oplsa_matrix(xval, phase):
    deg2rad = np.pi/180
    xval = xval*deg2rad
    if phase is None:
       phase = np.zeros(4)

    cos_0 = np.ones(len(xval))
    cos_1 =  (1 + np.cos(xval + phase[0] ) )
    cos_2 =  (1 - np.cos(2*xval + phase[1]))
    cos_3 =  (1 + np.cos(3*xval + phase[2]))
    cos_4 =  (1 - np.cos(4*xval + phase[3]))
    A_matrix = np.vstack((cos_0, cos_1, cos_2, cos_3, cos_4)).T
    # A_matrix = np.vstack((cos_1, cos_2, cos_3, cos_4)).T

    return A_matrix

def RyBe_matrix(xval, phase):
    deg2rad = np.pi/180
    xval = xval*deg2rad
    if phase is None:
       phase = np.zeros(6)

    cos_0 = np.ones(len(xval))
    cos_1 =  -np.cos(xval) 
    cos_2 = (np.cos(xval))**2
    cos_3 = -(np.cos(xval))**3 
    cos_4 = (np.cos(xval))**4
    cos_5 = -(np.cos(xval))**5
    A_matrix = np.vstack((cos_0, cos_1, cos_2, cos_3, cos_4, cos_5)).T

    return A_matrix

def get_rmsd(y_pred, y_data):
    MSE = np.square(np.subtract(y_pred, y_data)).mean()
    RMSD = np.sqrt(MSE)
    print('   MSE   = {:.5f} '.format(MSE) )
    print(' % RMSD  = {:.5f} '.format(RMSD*100) )

def get_fit(model, fname, coeffs, A_matrix, xval, qm_rel, mm_rel):
    if model == 0:                                   # 0 triggers on the OPLSA type 
       yfit = np.array(
                     coeffs[0]* A_matrix[:,0] +
                     coeffs[1]* A_matrix[:,1] +
                     coeffs[2]* A_matrix[:,2] +
                     coeffs[3]* A_matrix[:,3] +
                     coeffs[4]* A_matrix[:,4] +
                     mm_rel)
                    #  0.5* coeffs[0]* A_matrix[:,0] +
                    #  0.5* coeffs[1]* A_matrix[:,1] +
                    #  0.5* coeffs[2]* A_matrix[:,2] +
                    #  0.5* coeffs[3]* A_matrix[:,3] +
                    #  mm_rel)
    else:                                            # Otherwise Ryckart-Bellemans potential
        yfit = np.array( coeffs[0]* A_matrix[:,0] +
                         coeffs[1]* A_matrix[:,1] +
                         coeffs[2]* A_matrix[:,2] +
                         coeffs[3]* A_matrix[:,3] +
                         coeffs[4]* A_matrix[:,4] +
                         coeffs[5]* A_matrix[:,5] +
                         mm_rel)

    with open(fname, 'w') as f:
        for i, item in enumerate(yfit):
            f.write(f' {xval[i]}  {qm_rel[i]:.6f}  {mm_rel[i]:.6f}  {item:.6f} \n')
    f.close()

    return yfit
    
#####################
# NON LINEAR FITTING
#####################

def non_linear_f(x, a1, a2, a3, a4, p1, p2, p3, p4):
    y = a1*(1 + np.cos(x + p1)) + \
        a2*(1 - np.cos(2*x + p2)) + \
        a3*(1 - np.cos(3*x + p3)) + \
        a4*(1 - np.cos(4*x + p4)) 
    return y

def solve_LLSQ(A_matrix, yval):
    # coeffs = np.dot(np.linalg.inv(np.dot(A_matrix.T, A_matrix)), np.dot(A_matrix.T, yval))
    coeffs = np.linalg.lstsq(A_matrix, yval)[0]
    print(coeffs)
    # cov_matrix = np.diag(residual(coeffs, xval, yval)**2)
    print('', 60*'=')
    print(' Torsional Coefficients after LLSQ (kcal/mol): ')
    print('', 60*'=')
    [ print(' Coefficient # {} = {:.4f}'.format(i+1, x)) for i, x in enumerate(coeffs)]

    return coeffs


def rb2opls(coeffs):
    f4 = - coeffs[4] / 4
    f3 = - coeffs[3] / 2
    f2 = - coeffs[2] - coeffs[4] / 4
    f1 = - 2* coeffs[1] - (3/2)*coeffs[3]
    x = np.array([f1, f2, f3, f4])

    # A = np.array([[1/2, 1, 1/2, 0],
    #               [-1/2, 0, 3/2, 0],
    #               [0, -1, 0, 4],
    #               [0, 0, -2, 0]])
    # b = coeffs[:4]
    # x = np.linalg.solve(A, b)
    return x              
    # return c_opls

def main():
    parser = build_parser()
    args = parser.parse_args()
    fname = args.file
    data = load_data(fname)
    xval = data[:,0] # - 180.0
    qm_rel = ( data[:,1] - min(data[:,1]) ) *6.275030E02  # Eh to kcal/mol
    mm_rel = ( data[:,2] - min(data[:,2]) ) *6.275030E02
    yval =  ( qm_rel - mm_rel ) 
    pd.Series(qm_rel).to_csv('qm_rel.csv', index=False)
    pd.Series(mm_rel).to_csv('mm_rel.csv', index=False)
    
    A_oplsa = oplsa_matrix(xval, None)
    A_rybe = RyBe_matrix(xval, None)
    # LLSQ
    coeffs_oplsa = solve_LLSQ(A_oplsa, yval)
    coeffs_rybe = solve_LLSQ(A_rybe, yval)
    # c_oplsa_new = rb2opls(coeffs_rybe)
    # print(c_oplsa_new)
    fname_1 = 'oplsa_fitted.txt'
    fname_2 = 'rybe_fitted.txt'
    yfit_oplsa = get_fit(0, fname_1, coeffs_oplsa, A_oplsa, xval, qm_rel, mm_rel)
    yfit_rybe = get_fit(1, fname_2, coeffs_rybe, A_rybe, xval, qm_rel, mm_rel)
    # get_rmsd(yfit_oplsa, yval)
    # get_rmsd(yfit_rybe, yval)

    #  # WLLSQ 
    # kB = 8.617333262e-05
    # temp =  1000
    # kBT = kB * temp  # Boltzman constant in au at Room Temperature
    # # w = np.exp(-yval/kBT)
    # w = np.exp(-0.2 *np.sqrt(qm_rel))
    # print("")
    # print(w)
    # print("")

    # res = yval - np.dot(A_mat,coeffs)
    # # W_matrix = np.diag(w)
    # W_matrix = np.diag(res**2)
    # ATWI = np.dot(A_mat.T, np.linalg.inv(W_matrix))
    # prefac = np.dot(ATWI, A_mat)
    # fac = np.dot(ATWI, yval)
    # w_coeffs = np.dot(np.linalg.inv(prefac), fac)
    # print(w_coeffs)
    # print('', 60*'=')
    # print(' Torsional Coefficients after WLLSQ (kcal/mol): ')
    # print('', 60*'=')
    # [ print(' Coefficient # {} = {:.4f}'.format(i+1, x)) for i, x in enumerate(w_coeffs)]
    # fname = 'w_linear_fitted.txt'
    # A_mat_w = np.copy(A_mat)
    # yfit_w_linear = print_oplsTrs(fname, w_coeffs, A_mat_w, xval, qm_rel, mm_rel)
    # get_rmsd(yfit_w_linear, yval)


    # # NON-LLSQ
    # print('', 60*'=')
    # print(' Torsional Coefficients after Non-linear LLSQ (kcal/mol): ')
    # print('', 60*'=')
    # # p0 = [0.1, 0.1, 0.1, 0.1]
    # # p0 = np.ones(8)
    # p_0 = np.zeros(8)
    # # # popt, pcov = optimize.leastsq(residual, p0, args=(xval, yval))
    # popt, pcov = optimize.curve_fit(non_linear_f, xval, yval, p0=p_0)
    # # popt, pcov = optimize.curve_fit(residual, xval, yval)
    # [ print(' Coefficient # {} = {:.4f}'.format(i+1, x*6.275030E02)) for i, x in enumerate(popt)]
    # fname = 'non-linear_fitted.txt'
    # phase = popt[4:] * 180/np.pi
    # print(phase)
    # vs = popt[:4]
    # A_mat_shifted = oplstrs_matrix(xval, phase)
    # yfit_non_linear = print_oplsTrs(fname, vs, A_mat_shifted, xval, qm_rel, mm_rel)
    # get_rmsd(yfit_non_linear, yval)

    # # PLOT DATA
    if args.plot:
       fig = plt.subplots(1, figsize= (10,8), dpi=96 )
       plt.plot(xval, qm_rel, label='QM', marker='o', markersize=8, fillstyle='none', c='black', linestyle='dashed', dashes=(5, 10))
       plt.plot(xval, mm_rel, label='MM', marker='o', markersize=8, fillstyle='none', c='orangered', linestyle='dashed', dashes=(5, 10))
       plt.plot(xval, yfit_oplsa, label='Fitted$_{LLS}$', marker='o',  markersize=10, c='blue', alpha=0.6)
    #    plt.plot(xval, yfit_rybe, label='LLSQ$_{Ry-Be}$', marker='o',  markersize=10, c='green')
       # plt.plot(xval, yfit_w_linear, label='WLLSQ', marker='v')
       # plt.plot(xval, yfit_non_linear, label='NON-LLSQ', marker='v')
       plt.xlabel('Angle (°)', fontsize=18)
       plt.ylabel('$\Delta$E (kcal/mol)', fontsize=18)
       plt.xticks(fontsize=16)
       plt.yticks(fontsize=16)
       plt.legend(prop={'size':20})
       plt.tight_layout()
       plt.savefig(args.name +'.png')
       plt.show()



if __name__ == '__main__':
    main()
